"""
Example of how to import and use the brewblox service
"""

from brewblox_service import brewblox_logger, events, service

from brewblox_history import builder, influx, sse

LOGGER = brewblox_logger(__name__)


def create_parser(default_name='history'):
    parser = service.create_parser(default_name=default_name)
    parser.add_argument('--broadcast-exchange',
                        help='Eventbus exchange to which device services broadcast their state. [%(default)s]',
                        default='brewcast')
    parser.add_argument('--skip-influx-config',
                        help='Skip configuring Influx database after each connection. '
                        'This is useful when using custom configuration.',
                        action='store_true')
    return parser


def main():
    app = service.create_app(parser=create_parser())

    # Setup history functionality
    events.setup(app)
    influx.setup(app)
    builder.setup(app)
    sse.setup(app)

    influx.get_relay(app).subscribe(
        exchange_name=app['config']['broadcast_exchange'],
        routing='#'
    )

    service.furnish(app)
    service.run(app)


if __name__ == '__main__':
    main()



"""
Server-sent events implementation for relaying eventbus messages to front end
"""

import asyncio
import json
import time

from aiohttp import web
from aiohttp_sse import sse_response
from brewblox_history import builder, influx
from brewblox_service import brewblox_logger

LOGGER = brewblox_logger(__name__)
routes = web.RouteTableDef()

POLL_INTERVAL_S = 1


def setup(app: web.Application):
    app.router.add_routes(routes)


@routes.get('/sse/values')
async def subscribe(request: web.Request) -> web.Response:
    """
    ---
    tags:
    - History
    summary: subscribe to InfluxDB updates
    operationId: history.sse.values
    produces:
    - application/json
    parameters:
    -
        in: query
        name: database
        schema:
            type: string
            example: "brewblox"
            required: false
    -
        in: query
        name: measurement
        schema:
            type: string
            example: "spark"
            required: true
    -
        in: query
        name: keys
        schema:
            type: list
            required: false
            example: ["*"]
    """
    influx_client = influx.get_client(request.app)
    database = request.query.get('database')
    measurement = request.query.get('measurement')
    keys = []
    for k in request.query.getall('keys', ['*']):
        keys += k.split(',')

    async with sse_response(request) as resp:
        # Convert seconds to nanoseconds
        previous = int(time.time() * 1000 * 1000 * 1000)

        while True:
            try:
                await asyncio.sleep(POLL_INTERVAL_S)

                data = await builder.select_values(
                    influx_client,
                    database=database,
                    measurement=measurement,
                    keys=keys,
                    start=previous + 1
                )

                if data.get('values'):
                    await resp.send(json.dumps(data))
                    previous = data['values'][-1][0]

            except Exception as ex:
                LOGGER.warn(f'Exiting SSE with error: {ex}')
                break

    return resp



"""
Tests brewblox_history.__main__
"""

from brewblox_history import __main__ as main

TESTED = main.__name__


def test_main(mocker, sys_args):
    mocker.patch.object(main.service.sys, 'argv', sys_args)
    mocker.patch(TESTED + '.service.run')
    main.main()



"""
Tests brewblox_history.influx
"""

import asyncio
from unittest.mock import Mock, ANY

import pytest
from aiohttp.client_exceptions import ClientConnectionError
from asynctest import CoroutineMock

from brewblox_history import influx

TESTED = influx.__name__


class InfluxClientMock(Mock):

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass


@pytest.fixture
async def reduced_sleep(mocker):
    mocker.patch.object(influx, 'WRITE_INTERVAL_S', 0.001)
    mocker.patch.object(influx, 'RECONNECT_INTERVAL_S', 0.001)


@pytest.fixture
async def fewer_max_points(mocker):
    mocker.patch.object(influx, 'MAX_PENDING_POINTS', 10)


@pytest.fixture()
def influx_mock(mocker):
    m = InfluxClientMock()
    [setattr(m.return_value, f, CoroutineMock()) for f in [
        'ping',
        'create_database',
        'write',
        'query',
        'close'
    ]]

    mocker.patch(TESTED + '.InfluxDBClient', m)
    return m.return_value


@pytest.fixture
async def app(app, mocker, influx_mock, reduced_sleep):
    mocker.patch(TESTED + '.events.get_listener')

    influx.setup(app)
    return app


@pytest.fixture
async def app_skip_config(app):
    app['config']['skip_influx_config'] = True
    return app


async def test_setup(app, client):
    assert influx.get_writer(app)
    assert influx.get_relay(app)
    assert influx.get_client(app)


async def test_endpoints_offline(app, client):
    assert (await client.post('/subscribe', json={
        'exchange': 'brewblox',
        'routing': 'controller.#'
    })).status == 200


async def test_runtime_construction(app, client):
    # App is running, objects should still be createable
    query_client = influx.QueryClient()
    await query_client.startup(app)
    await query_client.shutdown()
    await query_client.shutdown()

    writer = influx.InfluxWriter()
    await writer.startup(app)
    assert writer.is_running
    await writer.shutdown()
    await writer.shutdown()

    relay = influx.EventRelay(app)
    await relay.startup(app)
    await relay.shutdown()
    await relay.shutdown()


async def test_query_client(app, client, influx_mock):
    query_client = influx.get_client(app)
    retval = dict(key='val')
    influx_mock.query.return_value = retval

    data = await query_client.query(database='db', query='gimme')

    assert data == retval


async def test_running_writer(influx_mock, app, client, mocker):
    writer = influx.get_writer(app)

    await writer.write_soon(
        measurement='measurement',
        fields=dict(field1=1, field2=2),
        tags=dict(tag1=1, tag2=2)
    )

    await asyncio.sleep(0.1)
    assert writer.is_running
    assert not writer._pending
    assert influx_mock.write.call_count == 1


async def test_run_error(influx_mock, app, client, mocker):
    writer = influx.get_writer(app)
    influx_mock.ping.side_effect = RuntimeError
    warn_mock = mocker.spy(influx.LOGGER, 'warn')

    await asyncio.sleep(0.1)

    assert not writer.is_running
    assert warn_mock.call_count == 1


async def test_retry_generate_connection(influx_mock, app, client):
    writer = influx.get_writer(app)
    await writer.shutdown()

    influx_mock.ping.reset_mock()
    influx_mock.create_database.reset_mock()

    influx_mock.ping.side_effect = ClientConnectionError

    await writer.startup(app)
    await asyncio.sleep(0.1)

    # generate_connections() keeps trying, but no success so far
    assert writer.is_running
    assert influx_mock.ping.call_count > 0
    assert influx_mock.create_database.call_count == 0


async def test_reconnect(influx_mock, app, client):
    writer = influx.get_writer(app)

    influx_mock.create_database.side_effect = ClientConnectionError

    await writer.shutdown()
    await writer.startup(app)

    await writer.write_soon(
        measurement='measurement',
        fields=dict(field1=1, field2=2),
        tags=dict(tag1=1, tag2=2)
    )

    await asyncio.sleep(0.1)
    assert influx_mock.write.call_count == 0
    assert writer.is_running

    influx_mock.create_database.side_effect = [True]

    await asyncio.sleep(0.1)
    assert writer.is_running
    assert influx_mock.write.call_count == 1


async def test_downsample(influx_mock, app, client, fewer_max_points):
    influx_mock.create_database.side_effect = ClientConnectionError

    writer = influx.get_writer(app)
    await writer.shutdown()
    await writer.startup(app)

    for i in range(2 * influx.MAX_PENDING_POINTS + 1):
        await writer.write_soon(
            measurement='measurement',
            fields=dict(field1=1, field2=2),
            tags=dict(tag1=1, tag2=2)
        )

    # It's been downsampled to half every time it hit max points
    assert len(writer._pending) == 0.5 * influx.MAX_PENDING_POINTS + 1
    assert influx_mock.write.call_count == 0
    influx_mock.create_database.side_effect = None
    await asyncio.sleep(0.1)


async def test_relay_subscribe(influx_mock, app, client):
    listener = influx.events.get_listener.return_value
    relay = influx.get_relay(app)
    relay.subscribe('arg', kw='kwarg')

    listener.subscribe.assert_called_once_with(
        'arg', kw='kwarg', on_message=relay._on_event_message)


async def test_relay_message(influx_mock, app, client):
    relay = influx.get_relay(app)

    data = {
        'nest': {
            'ed': {
                'values': [
                    'val',
                    'var'
                ]
            }
        }
    }

    flat_data = {
        'key/nest/ed/values/0': 'val',
        'key/nest/ed/values/1': 'var'
    }

    flat_value = {
        'single/text': 'value'
    }

    await relay._on_event_message(None, 'route.key', data)
    await relay._on_event_message(None, 'route.single', 'value')

    expected = [
        {'time': ANY, 'measurement': 'route', 'fields': flat_data, 'tags': {}},
        {'time': ANY, 'measurement': 'route', 'fields': flat_value, 'tags': {}}
    ]

    await asyncio.sleep(0.1)
    influx_mock.write.assert_called_once_with(expected)


async def test_skip_config(influx_mock, app_skip_config, client):
    await asyncio.sleep(0.1)
    assert influx_mock.create_database.call_count == 0



"""
Tests history.builder
"""

from unittest.mock import call

import pytest
from asynctest import CoroutineMock
from brewblox_history import builder

TESTED = builder.__name__


@pytest.fixture
def influx_mock(mocker):
    m = mocker.patch(TESTED + '.influx.get_client').return_value
    return m


@pytest.fixture
def query_mock(influx_mock):
    influx_mock.query = CoroutineMock(side_effect=lambda **kwargs: kwargs)
    return influx_mock.query


@pytest.fixture
async def app(app, mocker, query_mock):
    builder.setup(app)
    return app


@pytest.fixture
def field_keys_result():
    return {
        'results': [
            {
                'statement_id': 0,
                'series': [
                    {
                        'name': 'average_temperature',
                        'columns': [
                            'fieldKey',
                            'fieldType'
                        ],
                        'values': [
                            [
                                'degrees',
                                'float'
                            ]
                        ]
                    },
                    {
                        'name': 'h2o_feet',
                        'columns': [
                            'fieldKey',
                            'fieldType'
                        ],
                        'values': [
                            [
                                'level description',
                                'string'
                            ],
                            [
                                'water_level',
                                'float'
                            ]
                        ]
                    }
                ]
            }
        ]
    }


@pytest.fixture
def values_result():
    return {
        'results': [
            {
                'statement_id': 0,
                'series': [
                    {
                        'name': 'average_temperature',
                        'columns': [
                            'time',
                            'degrees',
                            'location'
                        ],
                        'values': [
                            [
                                1439856000000000000,
                                82,
                                'coyote_creek'
                            ],
                            [
                                1439856000000000000,
                                85,
                                'santa_monica'
                            ],
                            [
                                1439856360000000000,
                                73,
                                'coyote_creek'
                            ],
                            [
                                1439856360000000000,
                                74,
                                'santa_monica'
                            ],
                            [
                                1439856720000000000,
                                86,
                                'coyote_creek'
                            ],
                            [
                                1439856720000000000,
                                80,
                                'santa_monica'
                            ],
                            [
                                1439857080000000000,
                                89,
                                'coyote_creek'
                            ],
                            [
                                1439857080000000000,
                                81,
                                'santa_monica'
                            ],
                            [
                                1439857440000000000,
                                77,
                                'coyote_creek'
                            ],
                            [
                                1439857440000000000,
                                81,
                                'santa_monica'
                            ]
                        ]
                    }
                ]
            }
        ]
    }


async def test_custom_query(app, client, query_mock):
    content = {
        'database': 'brewblox',
        'query': 'select * from controller'
    }

    assert (await client.post('/_debug/query', json=content)).status == 200
    query_mock.assert_called_once_with(**content)


async def test_list_objects(app, client, query_mock, field_keys_result):
    query_mock.side_effect = lambda **kwargs: field_keys_result

    resp = await client.post('/query/objects', json={})
    resp_content = await resp.json()

    assert resp_content == {
        'average_temperature': [
            'degrees'
        ],
        'h2o_feet': [
            'level description',
            'water_level'
        ]
    }

    await client.post('/query/objects', json={'measurement': 'measy', 'injection': 'drop tables'})
    await client.post('/query/objects', json={'database': 'the_internet'})

    assert query_mock.mock_calls == [
        call(query='show field keys'),
        call(query='show field keys from {measurement}', measurement='measy'),
        call(database='the_internet', query='show field keys')
    ]


async def test_single_key(app, client, query_mock, values_result):
    """Asserts that ['single'] is split to 'single', and not 's,i,n,g,l,e'"""
    query_mock.side_effect = lambda **kwargs: values_result

    res = await client.post('/query/values', json={'measurement': 'm', 'keys': ['single']})
    assert res.status == 200

    query_mock.assert_called_once_with(
        query='select {keys} from {measurement}',
        measurement='m',
        keys='"single"'
    )


async def test_quote_keys(app, client, query_mock, values_result):
    """Keys must be quoted with double quotes. '*' is an exception."""
    query_mock.side_effect = lambda **kwargs: values_result

    res = await client.post('/query/values', json={'measurement': 'm', 'keys': ['first', 'second']})
    assert res.status == 200

    query_mock.assert_called_once_with(
        query='select {keys} from {measurement}',
        measurement='m',
        keys='"first","second"'
    )


async def test_value_data_format(app, client, query_mock, values_result):
    query_mock.side_effect = lambda **kwargs: values_result

    res = await client.post('/query/values', json={'measurement': 'm'})
    assert res.status == 200

    data = await res.json()
    assert data['name'] == 'average_temperature'
    assert len(data['values']) == 10
    assert data['columns'][1] == 'degrees'


@pytest.mark.parametrize('input_args, query_str', [
    (
        {},
        'select {keys} from {measurement}'
    ),
    (
        {'keys': ['you']},
        'select {keys} from {measurement}'
    ),
    (
        {'keys': ['key1', 'key2']},
        'select {keys} from {measurement}'
    ),
    (
        {'database': 'db'},
        'select {keys} from {measurement}'
    ),
    (
        {'start': 'a long time ago'},
        'select {keys} from {measurement} where time >= {start}'
    ),
    (
        {'start': 'once upon a time', 'duration': 'some time'},
        'select {keys} from {measurement} where time >= {start} and time <= {start} + {duration}'
    ),
    (
        {'start': 'then', 'end': 'now'},
        'select {keys} from {measurement} where time >= {start} and time <= {end}'
    ),
    (
        {'end': 'fire nation attack'},
        'select {keys} from {measurement} where time <= {end}'
    ),
    (
        {'end': 'life', 'duration': 'bright side'},
        'select {keys} from {measurement} where time >= {end} - {duration} and time <= {end}'
    ),
    (
        {'duration': 'eternal'},
        'select {keys} from {measurement} where time >= now() - {duration}'
    ),
    (
        {'keys': ['key1', 'key2'], 'order_by': 'time desc', 'limit': 1},
        'select {keys} from {measurement} order by {order_by} limit {limit}'
    ),
    (
        {'duration': 'eternal', 'limit': 1},
        'select {keys} from {measurement} where time >= now() - {duration} limit {limit}'
    ),
    (
        {'database': 'db', 'keys': ['something', 'else'], 'start': 'good old days', 'duration': '1d', 'limit': 5},
        'select {keys} from {measurement} where time >= {start} and time <= {start} + {duration} limit {limit}'
    )
])
async def test_get_values(input_args, query_str, app, client, query_mock, values_result):
    query_mock.side_effect = lambda **kwargs: values_result

    # Measurement is a required argument
    # Always add it to input_args
    input_args['measurement'] = input_args.get('measurement', 'emmy')

    call_args = input_args.copy()

    # Mirrors transformation in API:
    # * Query string is created
    # * Keys are converted from a list to a comma separated string
    call_args['query'] = query_str
    quoted_keys = [f'"{k}"' for k in input_args.get('keys', [])] or ['*']
    call_args['keys'] = ','.join(quoted_keys)

    res = await client.post('/query/values', json=input_args)
    assert res.status == 200
    query_mock.assert_called_once_with(**call_args)


async def test_invalid_time_frame(app, client):
    res = await client.post('/query/values', json={'measurement': 'm', 'start': 'x', 'duration': 'y', 'end': 'z'})
    assert res.status == 500
    assert 'ValueError' in await res.text()


async def test_no_values_found(app, client, query_mock):
    query_mock.side_effect = {'results': []}

    res = await client.post('/query/values', json={'measurement': 'm'})
    assert res.status == 200
    assert await res.json() == {}


async def test_error_response(app, client, query_mock):
    query_mock.side_effect = RuntimeError('Whoops.')
    resp = await client.post('/query/objects', json={})

    assert resp.status == 500
    assert 'Whoops.' in await resp.text()



"""
Converts REST endpoints into Influx queries
"""

from typing import Optional, Callable

import dpath
from aiohttp import web
from brewblox_history import influx
from brewblox_service import brewblox_logger

LOGGER = brewblox_logger(__name__)
routes = web.RouteTableDef()


def setup(app: web.Application):
    app.router.add_routes(routes)
    app.middlewares.append(controller_error_middleware)


@web.middleware
async def controller_error_middleware(request: web.Request, handler: web.RequestHandler) -> web.Response:
    try:
        return await handler(request)
    except Exception as ex:
        LOGGER.warn(f'REST error: {ex}')
        return web.json_response({'error': f'{type(ex).__name__}={ex}'}, status=500)


########################################################################################################


async def _do_with_handler(func: Callable, request: web.Request) -> web.Response:
    args = await request.json()
    response = await func(influx.get_client(request.app), **args)
    return web.json_response(response)


def _prune(vals: dict, relevant: set) -> dict:
    """Creates a dict only containing meaningful and relevant key/value pairs.

    All pairs in the returned dict met three conditions:
    * Their key was in `relevant`
    * Their key was in `vals`
    * The value associated with the key in `vals` was not None
    """
    return {k: vals[k] for k in relevant if vals.get(k) is not None}


def _find_time_frame(start: Optional[str], duration: Optional[str], end: Optional[str]) -> str:
    """
    Determines required InfluxDB QL where/and clause required to express time frame.

    A time frame can be constructed from a start, a duration, and an end.
    At most two determinators can be present simultaneously.
    """
    clause = ''

    if all([start, duration, end]):
        raise ValueError('At most two out of three duration arguments can be provided')

    elif not any([start, duration, end]):
        pass

    elif start and duration:
        clause = ' where time >= {start} and time <= {start} + {duration}'

    elif start and end:
        clause = ' where time >= {start} and time <= {end}'

    elif duration and end:
        clause = ' where time >= {end} - {duration} and time <= {end}'

    elif start:
        clause = ' where time >= {start}'

    elif duration:
        clause = ' where time >= now() - {duration}'

    elif end:
        clause = ' where time <= {end}'

    else:  # pragma: no cover
        # This path should never be reached
        raise RuntimeError('Unexpected code path while determining time frame!')

    return clause


########################################################################################################


async def raw_query(client: influx.QueryClient,
                    database: Optional[str]=influx.DEFAULT_DATABASE,
                    query: Optional[str]='show databases'
                    ) -> dict:

    return await client.query(
        query=query,
        database=database
    )


async def show_keys(client: influx.QueryClient,
                    database: Optional[str]=None,
                    measurement: Optional[str]=None,
                    **_  # allow, but discard all other kwargs
                    ) -> dict:

    query = 'show field keys'

    if measurement:
        query += ' from {measurement}'

    params = _prune(locals(), {'query', 'database', 'measurement'})
    query_response = await client.query(**params)

    response = dict()

    for path, meas_name in dpath.util.search(
            query_response, 'results/*/series/*/name', yielded=True, dirs=False):

        # results/[index]/series/[index]/values/*/0
        values_glob = '/'.join(path.split('/')[:-1] + ['values', '*', '0'])
        response[meas_name] = dpath.util.values(query_response, values_glob)

    return response


async def select_values(client: influx.QueryClient,
                        measurement: str,
                        keys: Optional[list]=['*'],
                        database: Optional[str]=None,
                        start: Optional[str]=None,
                        duration: Optional[str]=None,
                        end: Optional[str]=None,
                        order_by: Optional[str]=None,
                        limit: Optional[int]=None,
                        **_  # allow, but discard all other kwargs
                        ) -> dict:

    query = 'select {keys} from {measurement}'
    keys = ','.join([f'"{key}"' if key is not '*' else key for key in keys])

    query += _find_time_frame(start, duration, end)

    if order_by:
        query += ' order by {order_by}'

    if limit:
        query += ' limit {limit}'

    params = _prune(locals(), {'query', 'database', 'measurement', 'keys',
                               'start', 'duration', 'end', 'order_by', 'limit'})
    query_response = await client.query(**params)

    try:
        # Only support single-measurement queries
        response = dpath.util.get(query_response, 'results/0/series/0')
    except KeyError:
        # Nothing found
        response = dict()

    return response


########################################################################################################


@routes.post('/_debug/query')
async def custom_query(request: web.Request) -> web.Response:
    """
    ---
    tags:
    - History
    summary: Query InfluxDB
    description: Send a string query to the database.
    operationId: history.query
    produces:
    - application/json
    parameters:
    -
        in: body
        name: body
        description: Query
        required: true
        schema:
            type: object
            properties:
                database:
                    type: string
                query:
                    type: string
    """
    return await _do_with_handler(raw_query, request)


@routes.post('/query/objects')
async def objects_query(request: web.Request) -> web.Response:
    """
    ---
    tags:
    - History
    summary: List objects
    description: List available measurements and objects in database.
    operationId: history.query.objects
    produces:
    - application/json
    parameters:
    -
        in: body
        name: body
        required: true
        schema:
            type: object
            properties:
                database:
                    type: string
                    required: false
                measurement:
                    type: string
                    required: false
    """
    return await _do_with_handler(show_keys, request)


@routes.post('/query/values')
async def values_query(request: web.Request) -> web.Response:
    """
    ---
    tags:
    - History
    summary: Get object values
    operationId: history.query.values
    produces:
    - application/json
    parameters:
    -
        in: body
        name: body
        required: true
        schema:
            type: object
            properties:
                database:
                    type: string
                    required: false
                measurement:
                    type: string
                    required: true
                keys:
                    type: list
                    required: true
                    example: ["*"]
                start:
                    type: string
                    required: false
                    example: "1439873640000000000"
                duration:
                    type: string
                    required: false
                    example: "10m"
                end:
                    type: string
                    required: false
                    example: "1439873640000000000"
                limit:
                    type: int
                    required: false
                    example: 100
                order_by:
                    type: string
                    required: false
                    example: "time asc"
    """
    return await _do_with_handler(select_values, request)



"""
Tests brewblox_history.sse
"""

import json

import pytest
from asynctest import CoroutineMock
from brewblox_history import sse

TESTED = sse.__name__


@pytest.fixture
def values_result():
    return {
        'results': [
            {
                'statement_id': 0,
                'series': [
                    {
                        'name': 'average_temperature',
                        'columns': [
                            'time',
                            'degrees',
                            'location'
                        ],
                        'values': [
                            [
                                9000,
                                82,
                                'coyote_creek'
                            ],
                            [
                                9001,
                                85,
                                'santa_monica'
                            ]
                        ]
                    }
                ]
            }
        ]
    }


@pytest.fixture
def influx_mock(mocker):
    m = mocker.patch(TESTED + '.influx.get_client').return_value
    return m


@pytest.fixture
def interval_mock(mocker):
    mocker.patch(TESTED + '.POLL_INTERVAL_S', 0.001)


@pytest.fixture
async def app(app, influx_mock, interval_mock):
    sse.setup(app)
    return app


async def test_subscribe(app, client, influx_mock, values_result):
    influx_mock.query = CoroutineMock(side_effect=[{}, values_result, {}])
    res = await client.get('/sse/values', params={'measurement': 'm'})
    assert res.status == 200
    pushed = await res.text()
    assert json.loads(pushed[len('data:'):])  # SSE prefixes output with 'data: '
    assert influx_mock.query.call_count == 4
